@@ -167,8 +167,10 @@ void deinterleave_samples(const vector<float> &in, vector<float> *out_l, vector<
 
 }  // namespace
 
-AudioMixer::AudioMixer(unsigned num_cards)
-	: num_cards(num_cards),
+AudioMixer::AudioMixer(unsigned num_capture_cards, unsigned num_ffmpeg_inputs)
+	: num_capture_cards(num_capture_cards),
+	  num_ffmpeg_inputs(num_ffmpeg_inputs),
+	  ffmpeg_inputs(new AudioDevice[num_ffmpeg_inputs]),
 	  limiter(OUTPUT_FREQUENCY),
 	  correlation(OUTPUT_FREQUENCY)
 {
@@ -236,10 +238,8 @@ void AudioMixer::reset_resampler_mutex_held(DeviceSpec device_spec)
 	if (device->interesting_channels.empty()) {
 		device->resampling_queue.reset();
 	} else {
-		// TODO: ResamplingQueue should probably take the full device spec.
-		// (It's only used for console output, though.)
 		device->resampling_queue.reset(new ResamplingQueue(
-			device_spec.index, device->capture_frequency, OUTPUT_FREQUENCY, device->interesting_channels.size(),
+			device_spec, device->capture_frequency, OUTPUT_FREQUENCY, device->interesting_channels.size(),
 			global_flags.audio_queue_length_ms * 0.001));
 	}
 }
@@ -339,6 +339,7 @@ AudioMixer::BusSettings AudioMixer::get_default_bus_settings()
 	settings.fader_volume_db = 0.0f;
 	settings.muted = false;
 	settings.locut_enabled = global_flags.locut_enabled;
+	settings.stereo_width = 1.0f;
 	for (unsigned band_index = 0; band_index < NUM_EQ_BANDS; ++band_index) {
 		settings.eq_level_db[band_index] = 0.0f;
 	}
@@ -356,6 +357,7 @@ AudioMixer::BusSettings AudioMixer::get_bus_settings(unsigned bus_index) const
 	settings.fader_volume_db = fader_volume_db[bus_index];
 	settings.muted = mute[bus_index];
 	settings.locut_enabled = locut_enabled[bus_index];
+	settings.stereo_width = stereo_width[bus_index];
 	for (unsigned band_index = 0; band_index < NUM_EQ_BANDS; ++band_index) {
 		settings.eq_level_db[band_index] = eq_level_db[bus_index][band_index];
 	}
@@ -372,6 +374,7 @@ void AudioMixer::set_bus_settings(unsigned bus_index, const AudioMixer::BusSetti
 	fader_volume_db[bus_index] = settings.fader_volume_db;
 	mute[bus_index] = settings.muted;
 	locut_enabled[bus_index] = settings.locut_enabled;
+	stereo_width[bus_index] = settings.stereo_width;
 	for (unsigned band_index = 0; band_index < NUM_EQ_BANDS; ++band_index) {
 		eq_level_db[bus_index][band_index] = settings.eq_level_db[band_index];
 	}
@@ -389,6 +392,8 @@ AudioMixer::AudioDevice *AudioMixer::find_audio_device(DeviceSpec device)
 		return &video_cards[device.index];
 	case InputSourceType::ALSA_INPUT:
 		return &alsa_inputs[device.index];
+	case InputSourceType::FFMPEG_VIDEO_INPUT:
+		return &ffmpeg_inputs[device.index];
 	case InputSourceType::SILENCE:
 	default:
 		assert(false);
@@ -421,23 +426,60 @@ void AudioMixer::find_sample_src_from_device(const map<DeviceSpec, vector<float>
 }
 
 // TODO: Can be SSSE3-optimized if need be.
-void AudioMixer::fill_audio_bus(const map<DeviceSpec, vector<float>> &samples_card, const InputMapping::Bus &bus, unsigned num_samples, float *output)
+void AudioMixer::fill_audio_bus(const map<DeviceSpec, vector<float>> &samples_card, const InputMapping::Bus &bus, unsigned num_samples, float stereo_width, float *output)
 {
 	if (bus.device.type == InputSourceType::SILENCE) {
 		memset(output, 0, num_samples * 2 * sizeof(*output));
 	} else {
 		assert(bus.device.type == InputSourceType::CAPTURE_CARD ||
-		       bus.device.type == InputSourceType::ALSA_INPUT);
+		       bus.device.type == InputSourceType::ALSA_INPUT ||
+		       bus.device.type == InputSourceType::FFMPEG_VIDEO_INPUT);
 		const float *lsrc, *rsrc;
 		unsigned lstride, rstride;
 		float *dptr = output;
 		find_sample_src_from_device(samples_card, bus.device, bus.source_channel[0], &lsrc, &lstride);
 		find_sample_src_from_device(samples_card, bus.device, bus.source_channel[1], &rsrc, &rstride);
-		for (unsigned i = 0; i < num_samples; ++i) {
-			*dptr++ = *lsrc;
-			*dptr++ = *rsrc;
-			lsrc += lstride;
-			rsrc += rstride;
+
+		// Apply stereo width settings. Set stereo width w to a 0..1 range instead of
+		// -1..1, since it makes for much easier calculations (so 0.5 = completely mono).
+		// Then, what we want is
+		//
+		//   L' = wL + (1-w)R = R + w(L-R)
+		//   R' = wR + (1-w)L = L + w(R-L)
+		//
+		// This can be further simplified calculation-wise by defining the weighted
+		// difference signal D = w(R-L), so that:
+		//
+		//   L' = R - D
+		//   R' = L + D
+		float w = 0.5f * stereo_width + 0.5f;
+		if (bus.source_channel[0] == bus.source_channel[1]) {
+			// Mono anyway, so no need to bother.
+			w = 1.0f;
+		} else if (fabs(w) < 1e-3) {
+			// Perfect inverse.
+			swap(lsrc, rsrc);
+			swap(lstride, rstride);
+			w = 1.0f;
+		}
+		if (fabs(w - 1.0f) < 1e-3) {
+			// No calculations needed for stereo_width = 1.
+			for (unsigned i = 0; i < num_samples; ++i) {
+				*dptr++ = *lsrc;
+				*dptr++ = *rsrc;
+				lsrc += lstride;
+				rsrc += rstride;
+			}
+		} else {
+			// General case.
+			for (unsigned i = 0; i < num_samples; ++i) {
+				float left = *lsrc, right = *rsrc;
+				float diff = w * (right - left);
+				*dptr++ = right - diff;
+				*dptr++ = left + diff;
+				lsrc += lstride;
+				rsrc += rstride;
+			}
 		}
 	}
 }
@@ -457,6 +499,12 @@ vector<DeviceSpec> AudioMixer::get_active_devices() const
 			ret.push_back(device_spec);
 		}
 	}
+	for (unsigned card_index = 0; card_index < num_ffmpeg_inputs; ++card_index) {
+		const DeviceSpec device_spec{InputSourceType::FFMPEG_VIDEO_INPUT, card_index};
+		if (!find_audio_device(device_spec)->interesting_channels.empty()) {
+			ret.push_back(device_spec);
+		}
+	}
 	return ret;
 }
 
@@ -511,7 +559,7 @@ vector<float> AudioMixer::get_output(steady_clock::time_point ts, unsigned num_s
 	samples_out.resize(num_samples * 2);
 	samples_bus.resize(num_samples * 2);
 	for (unsigned bus_index = 0; bus_index < input_mapping.buses.size(); ++bus_index) {
-		fill_audio_bus(samples_card, input_mapping.buses[bus_index], num_samples, &samples_bus[0]);
+		fill_audio_bus(samples_card, input_mapping.buses[bus_index], num_samples, stereo_width[bus_index], &samples_bus[0]);
 		apply_eq(bus_index, &samples_bus);
 
 		{
@@ -880,7 +928,7 @@ map<DeviceSpec, DeviceInfo> AudioMixer::get_devices()
 	lock_guard<timed_mutex> lock(audio_mutex);
 
 	map<DeviceSpec, DeviceInfo> devices;
-	for (unsigned card_index = 0; card_index < num_cards; ++card_index) {
+	for (unsigned card_index = 0; card_index < num_capture_cards; ++card_index) {
 		const DeviceSpec spec{ InputSourceType::CAPTURE_CARD, card_index };
 		const AudioDevice *device = &video_cards[card_index];
 		DeviceInfo info;
@@ -900,6 +948,14 @@ map<DeviceSpec, DeviceInfo> AudioMixer::get_devices()
 		info.alsa_address = device.address;
 		devices.insert(make_pair(spec, info));
 	}
+	for (unsigned card_index = 0; card_index < num_ffmpeg_inputs; ++card_index) {
+		const DeviceSpec spec{ InputSourceType::FFMPEG_VIDEO_INPUT, card_index };
+		const AudioDevice *device = &ffmpeg_inputs[card_index];
+		DeviceInfo info;
+		info.display_name = device->display_name;
+		info.num_channels = 2;
+		devices.insert(make_pair(spec, info));
+	}
 	return devices;
 }
 
@@ -926,16 +982,25 @@ void AudioMixer::serialize_device(DeviceSpec device_spec, DeviceSpecProto *devic
 		case InputSourceType::ALSA_INPUT:
 			alsa_pool.serialize_device(device_spec.index, device_spec_proto);
 			break;
+		case InputSourceType::FFMPEG_VIDEO_INPUT:
+			device_spec_proto->set_type(DeviceSpecProto::FFMPEG_VIDEO_INPUT);
+			device_spec_proto->set_index(device_spec.index);
+			device_spec_proto->set_display_name(ffmpeg_inputs[device_spec.index].display_name);
+			break;
 	}
 }
 
 void AudioMixer::set_simple_input(unsigned card_index)
 {
+	assert(card_index < num_capture_cards + num_ffmpeg_inputs);
 	InputMapping new_input_mapping;
 	InputMapping::Bus input;
 	input.name = "Main";
-	input.device.type = InputSourceType::CAPTURE_CARD;
-	input.device.index = card_index;
+	if (card_index >= num_capture_cards) {
+		input.device = DeviceSpec{InputSourceType::FFMPEG_VIDEO_INPUT, card_index - num_capture_cards};
+	} else {
+		input.device = DeviceSpec{InputSourceType::CAPTURE_CARD, card_index};
+	}
 	input.source_channel[0] = 0;
 	input.source_channel[1] = 1;
 
@@ -955,6 +1020,11 @@ unsigned AudioMixer::get_simple_input() const
 	    input_mapping.buses[0].source_channel[0] == 0 &&
 	    input_mapping.buses[0].source_channel[1] == 1) {
 		return input_mapping.buses[0].device.index;
+	} else if (input_mapping.buses.size() == 1 &&
+	           input_mapping.buses[0].device.type == InputSourceType::FFMPEG_VIDEO_INPUT &&
+	           input_mapping.buses[0].source_channel[0] == 0 &&
+	           input_mapping.buses[0].source_channel[1] == 1) {
+		return input_mapping.buses[0].device.index + num_capture_cards;
 	} else {
 		return numeric_limits<unsigned>::max();
 	}
@@ -978,12 +1048,15 @@ void AudioMixer::set_input_mapping_lock_held(const InputMapping &new_input_mappi
 	map<DeviceSpec, set<unsigned>> interesting_channels;
 	for (const InputMapping::Bus &bus : new_input_mapping.buses) {
 		if (bus.device.type == InputSourceType::CAPTURE_CARD ||
-		    bus.device.type == InputSourceType::ALSA_INPUT) {
+		    bus.device.type == InputSourceType::ALSA_INPUT ||
+		    bus.device.type == InputSourceType::FFMPEG_VIDEO_INPUT) {
 			for (unsigned channel = 0; channel < 2; ++channel) {
 				if (bus.source_channel[channel] != -1) {
 					interesting_channels[bus.device].insert(bus.source_channel[channel]);
 				}
 			}
+		} else {
+			assert(bus.device.type == InputSourceType::SILENCE);
 		}
 	}
 
@@ -1023,6 +1096,8 @@ void AudioMixer::set_input_mapping_lock_held(const InputMapping &new_input_mappi
 			metrics.labels.emplace_back("source_type", "capture_card");
 		} else if (bus.device.type == InputSourceType::ALSA_INPUT) {
 			metrics.labels.emplace_back("source_type", "alsa_input");
+		} else if (bus.device.type == InputSourceType::FFMPEG_VIDEO_INPUT) {
+			metrics.labels.emplace_back("source_type", "ffmpeg_video_input");
 		} else {
 			assert(false);
 		}
@@ -1066,6 +1141,14 @@ void AudioMixer::set_input_mapping_lock_held(const InputMapping &new_input_mappi
 			reset_resampler_mutex_held(device_spec);
 		}
 	}
+	for (unsigned card_index = 0; card_index < num_ffmpeg_inputs; ++card_index) {
+		const DeviceSpec device_spec{InputSourceType::FFMPEG_VIDEO_INPUT, card_index};
+		AudioDevice *device = find_audio_device(device_spec);
+		if (device->interesting_channels != interesting_channels[device_spec]) {
+			device->interesting_channels = interesting_channels[device_spec];
+			reset_resampler_mutex_held(device_spec);
+		}
+	}
 
 	input_mapping = new_input_mapping;
 }
@@ -1095,4 +1178,18 @@ void AudioMixer::reset_peak(unsigned bus_index)
 	}
 }
 
+bool AudioMixer::is_mono(unsigned bus_index)
+{
+	lock_guard<timed_mutex> lock(audio_mutex);
+	const InputMapping::Bus &bus = input_mapping.buses[bus_index];
+	if (bus.device.type == InputSourceType::SILENCE) {
+		return true;
+	} else {
+		assert(bus.device.type == InputSourceType::CAPTURE_CARD ||
+		       bus.device.type == InputSourceType::ALSA_INPUT ||
+		       bus.device.type == InputSourceType::FFMPEG_VIDEO_INPUT);
+		return bus.source_channel[0] == bus.source_channel[1];
+	}
+}
+
 AudioMixer *global_audio_mixer = nullptr;
