@@ -208,9 +208,10 @@ YCbCrFormat decode_ycbcr_format(const AVPixFmtDescriptor *desc, const AVFrame *f
 FFmpegCapture::FFmpegCapture(const string &filename, unsigned width, unsigned height)
 	: filename(filename), width(width), height(height), video_timebase{1, 1}
 {
-	// Not really used for anything.
 	description = "Video: " + filename;
 
+	last_frame = steady_clock::now();
+
 	avformat_network_init();  // In case someone wants this.
 }
 
@@ -281,13 +282,20 @@ void FFmpegCapture::producer_thread_func()
 	pthread_setname_np(pthread_self(), thread_name);
 
 	while (!producer_thread_should_quit.should_quit()) {
-		string pathname = search_for_file(filename);
-		if (filename.empty()) {
-			fprintf(stderr, "%s not found, sleeping one second and trying again...\n", filename.c_str());
+		string filename_copy;
+		{
+			lock_guard<mutex> lock(filename_mu);
+			filename_copy = filename;
+		}
+
+		string pathname = search_for_file(filename_copy);
+		if (pathname.empty()) {
+			fprintf(stderr, "%s not found, sleeping one second and trying again...\n", filename_copy.c_str());
 			send_disconnected_frame();
 			producer_thread_should_quit.sleep_for(seconds(1));
 			continue;
 		}
+		should_interrupt = false;
 		if (!play_video(pathname)) {
 			// Error.
 			fprintf(stderr, "Error when playing %s, sleeping one second and trying again...\n", pathname.c_str());
@@ -319,19 +327,27 @@ void FFmpegCapture::send_disconnected_frame()
 		if (pixel_format == bmusb::PixelFormat_8BitBGRA) {
 			video_format.stride = width * 4;
 			video_frame.len = width * height * 4;
+			memset(video_frame.data, 0, video_frame.len);
 		} else {
 			video_format.stride = width;
+			current_frame_ycbcr_format.luma_coefficients = YCBCR_REC_709;
 			current_frame_ycbcr_format.full_range = true;
 			current_frame_ycbcr_format.num_levels = 256;
 			current_frame_ycbcr_format.chroma_subsampling_x = 2;
 			current_frame_ycbcr_format.chroma_subsampling_y = 2;
+			current_frame_ycbcr_format.cb_x_position = 0.0f;
+			current_frame_ycbcr_format.cb_y_position = 0.0f;
+			current_frame_ycbcr_format.cr_x_position = 0.0f;
+			current_frame_ycbcr_format.cr_y_position = 0.0f;
 			video_frame.len = width * height * 2;
+			memset(video_frame.data, 0, width * height);
+			memset(video_frame.data + width * height, 128, width * height);  // Valid for both NV12 and planar.
 		}
-		memset(video_frame.data, 0, video_frame.len);
 
 		frame_callback(-1, AVRational{1, TIMEBASE}, -1, AVRational{1, TIMEBASE}, timecode++,
 			video_frame, /*video_offset=*/0, video_format,
 			FrameAllocator::Frame(), /*audio_offset=*/0, AudioFormat());
+		last_frame_was_connected = false;
 	}
 }
 
@@ -349,7 +365,10 @@ bool FFmpegCapture::play_video(const string &pathname)
 		last_modified = buf.st_mtim;
 	}
 
-	auto format_ctx = avformat_open_input_unique(pathname.c_str(), nullptr, nullptr);
+	AVDictionary *opts = nullptr;
+	av_dict_set(&opts, "fflags", "nobuffer", 0);
+
+	auto format_ctx = avformat_open_input_unique(pathname.c_str(), nullptr, &opts, AVIOInterruptCB{ &FFmpegCapture::interrupt_cb_thunk, this });
 	if (format_ctx == nullptr) {
 		fprintf(stderr, "%s: Error opening file\n", pathname.c_str());
 		return false;
@@ -414,6 +433,7 @@ bool FFmpegCapture::play_video(const string &pathname)
 	internal_rewind();
 
 	// Main loop.
+	bool first_frame = true;
 	while (!producer_thread_should_quit.should_quit()) {
 		if (process_queued_commands(format_ctx.get(), pathname, last_modified, /*rewound=*/nullptr)) {
 			return true;
@@ -462,15 +482,72 @@ bool FFmpegCapture::play_video(const string &pathname)
 				pts_origin = frame->pts;	
 			}
 			next_frame_start = compute_frame_start(frame->pts, pts_origin, video_timebase, start, rate);
+			if (first_frame && last_frame_was_connected) {
+				// If reconnect took more than one second, this is probably a live feed,
+				// and we should reset the resampler. (Or the rate is really, really low,
+				// in which case a reset on the first frame is fine anyway.)
+				if (duration<double>(next_frame_start - last_frame).count() >= 1.0) {
+					last_frame_was_connected = false;
+				}
+			}
 			video_frame->received_timestamp = next_frame_start;
+
+			// The easiest way to get all the rate conversions etc. right is to move the
+			// audio PTS into the video PTS timebase and go from there. (We'll get some
+			// rounding issues, but they should not be a big problem.)
+			int64_t audio_pts_as_video_pts = av_rescale_q(audio_pts, audio_timebase, video_timebase);
+			audio_frame->received_timestamp = compute_frame_start(audio_pts_as_video_pts, pts_origin, video_timebase, start, rate);
+
+			if (audio_frame->len != 0) {
+				// The received timestamps in Nageru are measured after we've just received the frame.
+				// However, pts (especially audio pts) is at the _beginning_ of the frame.
+				// If we have locked audio, the distinction doesn't really matter, as pts is
+				// on a relative scale and a fixed offset is fine. But if we don't, we will have
+				// a different number of samples each time, which will cause huge audio jitter
+				// and throw off the resampler.
+				//
+				// In a sense, we should have compensated by adding the frame and audio lengths
+				// to video_frame->received_timestamp and audio_frame->received_timestamp respectively,
+				// but that would mean extra waiting in sleep_until(). All we need is that they
+				// are correct relative to each other, though (and to the other frames we send),
+				// so just align the end of the audio frame, and we're fine.
+				size_t num_samples = (audio_frame->len * 8) / audio_format.bits_per_sample / audio_format.num_channels;
+				double offset = double(num_samples) / OUTPUT_FREQUENCY -
+					double(video_format.frame_rate_den) / video_format.frame_rate_nom;
+				audio_frame->received_timestamp += duration_cast<steady_clock::duration>(duration<double>(offset));
+			}
+
+			steady_clock::time_point now = steady_clock::now();
+			if (duration<double>(now - next_frame_start).count() >= 0.1) {
+				// If we don't have enough CPU to keep up, or if we have a live stream
+				// where the initial origin was somehow wrong, we could be behind indefinitely.
+				// In particular, this will give the audio resampler problems as it tries
+				// to speed up to reduce the delay, hitting the low end of the buffer every time.
+				fprintf(stderr, "%s: Playback %.0f ms behind, resetting time scale\n",
+					pathname.c_str(),
+					1e3 * duration<double>(now - next_frame_start).count());
+				pts_origin = frame->pts;
+				start = next_frame_start = now;
+				timecode += MAX_FPS * 2 + 1;
+			}
 			bool finished_wakeup = producer_thread_should_quit.sleep_until(next_frame_start);
 			if (finished_wakeup) {
 				if (audio_frame->len > 0) {
 					assert(audio_pts != -1);
 				}
+				if (!last_frame_was_connected) {
+					// We're recovering from an error (or really slow load, see above).
+					// Make sure to get the audio resampler reset. (This is a hack;
+					// ideally, the frame callback should just accept a way to signal
+					// audio discontinuity.)
+					timecode += MAX_FPS * 2 + 1;
+				}
 				frame_callback(frame->pts, video_timebase, audio_pts, audio_timebase, timecode++,
 					video_frame.get_and_release(), 0, video_format,
 					audio_frame.get_and_release(), 0, audio_format);
+				first_frame = false;
+				last_frame = steady_clock::now();
+				last_frame_was_connected = true;
 				break;
 			} else {
 				if (producer_thread_should_quit.should_quit()) break;
@@ -559,6 +636,7 @@ AVFrameWithDeleter FFmpegCapture::decode_frame(AVFormatContext *format_ctx, AVCo
 	AVFrameWithDeleter video_avframe = av_frame_alloc_unique();
 	bool eof = false;
 	*audio_pts = -1;
+	bool has_audio = false;
 	do {
 		AVPacket pkt;
 		unique_ptr<AVPacket, decltype(av_packet_unref)*> pkt_cleanup(
@@ -577,9 +655,7 @@ AVFrameWithDeleter FFmpegCapture::decode_frame(AVFormatContext *format_ctx, AVCo
 					return AVFrameWithDeleter(nullptr);
 				}
 			} else if (pkt.stream_index == audio_stream_index) {
-				if (*audio_pts == -1) {
-					*audio_pts = pkt.pts;
-				}
+				has_audio = true;
 				if (avcodec_send_packet(audio_codec_ctx, &pkt) < 0) {
 					fprintf(stderr, "%s: Cannot send packet to audio codec.\n", pathname.c_str());
 					*error = true;
@@ -591,10 +667,13 @@ AVFrameWithDeleter FFmpegCapture::decode_frame(AVFormatContext *format_ctx, AVCo
 		}
 
 		// Decode audio, if any.
-		if (*audio_pts != -1) {
+		if (has_audio) {
 			for ( ;; ) {
 				int err = avcodec_receive_frame(audio_codec_ctx, audio_avframe.get());
 				if (err == 0) {
+					if (*audio_pts == -1) {
+						*audio_pts = audio_avframe->pts;
+					}
 					convert_audio(audio_avframe.get(), audio_frame, audio_format);
 				} else if (err == AVERROR(EAGAIN)) {
 					break;
@@ -672,7 +751,7 @@ void FFmpegCapture::convert_audio(const AVFrame *audio_avframe, FrameAllocator::
 		}
 
 		av_opt_set_int(resampler, "in_channel_layout",  channel_layout,                             0);
-		av_opt_set_int(resampler, "out_channel_layout", AV_CH_LAYOUT_STEREO,                        0);
+		av_opt_set_int(resampler, "out_channel_layout", AV_CH_LAYOUT_STEREO_DOWNMIX,                0);
 		av_opt_set_int(resampler, "in_sample_rate",     av_frame_get_sample_rate(audio_avframe),    0);
 		av_opt_set_int(resampler, "out_sample_rate",    OUTPUT_FREQUENCY,                           0);
 		av_opt_set_int(resampler, "in_sample_fmt",      audio_avframe->format,                      0);
@@ -797,3 +876,13 @@ UniqueFrame FFmpegCapture::make_video_frame(const AVFrame *frame, const string &
 
 	return video_frame;
 }
+
+int FFmpegCapture::interrupt_cb_thunk(void *unique)
+{
+	return reinterpret_cast<FFmpegCapture *>(unique)->interrupt_cb();
+}
+
+int FFmpegCapture::interrupt_cb()
+{
+	return should_interrupt.load();
+}
