@@ -8,6 +8,7 @@
 #include <stdint.h>
 #include <QAction>
 #include <QActionGroup>
+#include <QInputDialog>
 #include <QList>
 #include <QMenu>
 #include <QPoint>
@@ -32,11 +33,23 @@ class QMouseEvent;
 #include <movit/util.h>
 #include <string>
 
-
 using namespace movit;
 using namespace std;
 using namespace std::placeholders;
 
+namespace {
+
+double srgb_to_linear(double x)
+{
+	if (x < 0.04045) {
+		return x / 12.92;
+	} else {
+		return pow((x + 0.055) / 1.055, 2.4);
+	}
+}
+
+}  // namespace
+
 GLWidget::GLWidget(QWidget *parent)
     : QGLWidget(parent, global_share_widget)
 {
@@ -55,6 +68,22 @@ void GLWidget::shutdown()
 	global_mixer->remove_frame_ready_callback(output, this);
 }
 
+void GLWidget::grab_white_balance(unsigned channel, unsigned x, unsigned y)
+{
+	// Set the white balance to neutral for the grab. It's probably going to
+	// flicker a bit, but hopefully this display is not live anyway.
+	global_mixer->set_wb(output, 0.5, 0.5, 0.5);
+	global_mixer->wait_for_next_frame();
+
+	// Mark that the next paintGL() should grab the given pixel.
+	grab_x = x;
+	grab_y = y;
+	grab_output = Mixer::Output(Mixer::OUTPUT_INPUT0 + channel);
+	should_grab = true;
+
+	updateGL();
+}
+
 void GLWidget::initializeGL()
 {
 	static once_flag flag;
@@ -121,6 +150,17 @@ void GLWidget::paintGL()
 	} else {
 		assert(resource_pool == frame.chain->get_resource_pool());
 	}
+
+	if (should_grab) {
+		GLfloat reference_color[4];
+		glReadPixels(grab_x, current_height - grab_y - 1, 1, 1, GL_BGRA, GL_FLOAT, reference_color);
+
+		double r = srgb_to_linear(reference_color[2]);
+		double g = srgb_to_linear(reference_color[1]);
+		double b = srgb_to_linear(reference_color[0]);
+		global_mixer->set_wb(grab_output, r, g, b);
+		should_grab = false;
+	}
 }
 
 void GLWidget::mousePressEvent(QMouseEvent *event)
@@ -175,136 +215,152 @@ void GLWidget::show_preview_context_menu(unsigned signal_num, const QPoint &pos)
 	QMenu card_submenu;
 	QActionGroup card_group(&card_submenu);
 
-	unsigned num_cards = global_mixer->get_num_cards();
-	unsigned current_card = global_mixer->map_signal(signal_num);
-	for (unsigned card_index = 0; card_index < num_cards; ++card_index) {
-		QString description(QString::fromStdString(global_mixer->get_card_description(card_index)));
-		QAction *action = new QAction(description, &card_group);
-		action->setCheckable(true);
-		if (current_card == card_index) {
-			action->setChecked(true);
-		}
-		action->setData(QList<QVariant>{"card", card_index});
-		card_submenu.addAction(action);
-	}
-
-	card_submenu.setTitle("Input source");
-	menu.addMenu(&card_submenu);
-
-	// Note that this setting depends on which card is active.
-	// TODO: Consider hiding this for BGRA sources.
-
 	QMenu interpretation_submenu;
 	QActionGroup interpretation_group(&interpretation_submenu);
 
-	YCbCrInterpretation current_interpretation = global_mixer->get_input_ycbcr_interpretation(current_card);
-	{
-		QAction *action = new QAction("Auto", &interpretation_group);
-		action->setCheckable(true);
-		if (current_interpretation.ycbcr_coefficients_auto) {
-			action->setChecked(true);
-		}
-		action->setData(QList<QVariant>{"interpretation", true, YCBCR_REC_709, false});
-		interpretation_submenu.addAction(action);
-	}
-	for (YCbCrLumaCoefficients ycbcr_coefficients : { YCBCR_REC_709, YCBCR_REC_601 }) {
-		for (bool full_range : { false, true }) {
-			std::string description;
-			if (ycbcr_coefficients == YCBCR_REC_709) {
-				description = "Rec. 709 (HD)";
-			} else {
-				description = "Rec. 601 (SD)";
-			}
-			if (full_range) {
-				description += ", full range (nonstandard)";
-			}
-			QAction *action = new QAction(QString::fromStdString(description), &interpretation_group);
+	QMenu video_input_submenu;
+	QActionGroup video_input_group(&video_input_submenu);
+
+	QMenu audio_input_submenu;
+	QActionGroup audio_input_group(&audio_input_submenu);
+
+	QMenu mode_submenu;
+	QActionGroup mode_group(&mode_submenu);
+
+	unsigned num_cards = global_mixer->get_num_cards();
+	unsigned current_card = global_mixer->map_signal(signal_num);
+	bool is_ffmpeg = global_mixer->card_is_ffmpeg(current_card);
+
+	if (!is_ffmpeg) {  // FFmpeg inputs are not connected to any card; they're locked to a given input and have a given Y'CbCr interpretatio and have a given Y'CbCr interpretationn.
+		for (unsigned card_index = 0; card_index < num_cards; ++card_index) {
+			QString description(QString::fromStdString(global_mixer->get_card_description(card_index)));
+			QAction *action = new QAction(description, &card_group);
 			action->setCheckable(true);
-			if (!current_interpretation.ycbcr_coefficients_auto &&
-			    ycbcr_coefficients == current_interpretation.ycbcr_coefficients &&
-			    full_range == current_interpretation.full_range) {
+			if (current_card == card_index) {
 				action->setChecked(true);
 			}
-			action->setData(QList<QVariant>{"interpretation", false, ycbcr_coefficients, full_range});
+			action->setData(QList<QVariant>{"card", card_index});
+			card_submenu.addAction(action);
+		}
+
+		card_submenu.setTitle("Input source");
+		menu.addMenu(&card_submenu);
+
+		// Note that this setting depends on which card is active.
+
+		YCbCrInterpretation current_interpretation = global_mixer->get_input_ycbcr_interpretation(current_card);
+		{
+			QAction *action = new QAction("Auto", &interpretation_group);
+			action->setCheckable(true);
+			if (current_interpretation.ycbcr_coefficients_auto) {
+				action->setChecked(true);
+			}
+			action->setData(QList<QVariant>{"interpretation", true, YCBCR_REC_709, false});
 			interpretation_submenu.addAction(action);
 		}
-	}
+		for (YCbCrLumaCoefficients ycbcr_coefficients : { YCBCR_REC_709, YCBCR_REC_601 }) {
+			for (bool full_range : { false, true }) {
+				std::string description;
+				if (ycbcr_coefficients == YCBCR_REC_709) {
+					description = "Rec. 709 (HD)";
+				} else {
+					description = "Rec. 601 (SD)";
+				}
+				if (full_range) {
+					description += ", full range (nonstandard)";
+				}
+				QAction *action = new QAction(QString::fromStdString(description), &interpretation_group);
+				action->setCheckable(true);
+				if (!current_interpretation.ycbcr_coefficients_auto &&
+				    ycbcr_coefficients == current_interpretation.ycbcr_coefficients &&
+				    full_range == current_interpretation.full_range) {
+					action->setChecked(true);
+				}
+				action->setData(QList<QVariant>{"interpretation", false, ycbcr_coefficients, full_range});
+				interpretation_submenu.addAction(action);
+			}
+		}
 
-	interpretation_submenu.setTitle("Input interpretation");
-	menu.addMenu(&interpretation_submenu);
+		interpretation_submenu.setTitle("Input interpretation");
+		menu.addMenu(&interpretation_submenu);
+	}
 
 	// --- The choices in the next few options depend a lot on which card is active ---
 
-	// Add a submenu for selecting video input, with an action for each input.
-	QMenu video_input_submenu;
-	QActionGroup video_input_group(&video_input_submenu);
-	std::map<uint32_t, string> video_inputs = global_mixer->get_available_video_inputs(current_card);
-	uint32_t current_video_input = global_mixer->get_current_video_input(current_card);
-	for (const auto &mode : video_inputs) {
-		QString description(QString::fromStdString(mode.second));
-		QAction *action = new QAction(description, &video_input_group);
-		action->setCheckable(true);
-		if (mode.first == current_video_input) {
-			action->setChecked(true);
-		}
-		action->setData(QList<QVariant>{"video_input", mode.first});
-		video_input_submenu.addAction(action);
-	}
-
-	video_input_submenu.setTitle("Video input");
-	menu.addMenu(&video_input_submenu);
-
-	// The same for audio input.
-	QMenu audio_input_submenu;
-	QActionGroup audio_input_group(&audio_input_submenu);
-	std::map<uint32_t, string> audio_inputs = global_mixer->get_available_audio_inputs(current_card);
-	uint32_t current_audio_input = global_mixer->get_current_audio_input(current_card);
-	for (const auto &mode : audio_inputs) {
-		QString description(QString::fromStdString(mode.second));
-		QAction *action = new QAction(description, &audio_input_group);
-		action->setCheckable(true);
-		if (mode.first == current_audio_input) {
-			action->setChecked(true);
-		}
-		action->setData(QList<QVariant>{"audio_input", mode.first});
-		audio_input_submenu.addAction(action);
-	}
-
-	audio_input_submenu.setTitle("Audio input");
-	menu.addMenu(&audio_input_submenu);
-
-	// The same for resolution.
-	QMenu mode_submenu;
-	QActionGroup mode_group(&mode_submenu);
-	std::map<uint32_t, bmusb::VideoMode> video_modes = global_mixer->get_available_video_modes(current_card);
-	uint32_t current_video_mode = global_mixer->get_current_video_mode(current_card);
 	bool has_auto_mode = false;
-	for (const auto &mode : video_modes) {
-		QString description(QString::fromStdString(mode.second.name));
-		QAction *action = new QAction(description, &mode_group);
-		action->setCheckable(true);
-		if (mode.first == current_video_mode) {
-			action->setChecked(true);
+	QAction *change_url_action = nullptr;
+	if (is_ffmpeg) {
+		// Add a menu to change the source URL if we're an FFmpeg card.
+		// (The theme can still override.)
+		if (global_mixer->card_is_ffmpeg(current_card)) {
+			change_url_action = new QAction("Change source filename/URL…", &menu);
+			menu.addAction(change_url_action);
 		}
-		action->setData(QList<QVariant>{"video_mode", mode.first});
-		mode_submenu.addAction(action);
-
-		// TODO: Relying on the 0 value here (from bmusb.h) is ugly, it should be a named constant.
-		if (mode.first == 0) {
-			has_auto_mode = true;
+	} else {
+		// Add a submenu for selecting video input, with an action for each input.
+		std::map<uint32_t, string> video_inputs = global_mixer->get_available_video_inputs(current_card);
+		uint32_t current_video_input = global_mixer->get_current_video_input(current_card);
+		for (const auto &mode : video_inputs) {
+			QString description(QString::fromStdString(mode.second));
+			QAction *action = new QAction(description, &video_input_group);
+			action->setCheckable(true);
+			if (mode.first == current_video_input) {
+				action->setChecked(true);
+			}
+			action->setData(QList<QVariant>{"video_input", mode.first});
+			video_input_submenu.addAction(action);
 		}
-	}
 
-	// Add a “scan” menu if there's no “auto” mode.
-	if (!has_auto_mode) {
-		QAction *action = new QAction("Scan", &mode_group);
-		action->setData(QList<QVariant>{"video_mode", 0});
-		mode_submenu.addSeparator();
-		mode_submenu.addAction(action);
-	}
+		video_input_submenu.setTitle("Video input");
+		menu.addMenu(&video_input_submenu);
 
-	mode_submenu.setTitle("Input mode");
-	menu.addMenu(&mode_submenu);
+		// The same for audio input.
+		std::map<uint32_t, string> audio_inputs = global_mixer->get_available_audio_inputs(current_card);
+		uint32_t current_audio_input = global_mixer->get_current_audio_input(current_card);
+		for (const auto &mode : audio_inputs) {
+			QString description(QString::fromStdString(mode.second));
+			QAction *action = new QAction(description, &audio_input_group);
+			action->setCheckable(true);
+			if (mode.first == current_audio_input) {
+				action->setChecked(true);
+			}
+			action->setData(QList<QVariant>{"audio_input", mode.first});
+			audio_input_submenu.addAction(action);
+		}
+
+		audio_input_submenu.setTitle("Audio input");
+		menu.addMenu(&audio_input_submenu);
+
+		// The same for resolution.
+		std::map<uint32_t, bmusb::VideoMode> video_modes = global_mixer->get_available_video_modes(current_card);
+		uint32_t current_video_mode = global_mixer->get_current_video_mode(current_card);
+		for (const auto &mode : video_modes) {
+			QString description(QString::fromStdString(mode.second.name));
+			QAction *action = new QAction(description, &mode_group);
+			action->setCheckable(true);
+			if (mode.first == current_video_mode) {
+				action->setChecked(true);
+			}
+			action->setData(QList<QVariant>{"video_mode", mode.first});
+			mode_submenu.addAction(action);
+
+			// TODO: Relying on the 0 value here (from bmusb.h) is ugly, it should be a named constant.
+			if (mode.first == 0) {
+				has_auto_mode = true;
+			}
+		}
+
+		// Add a “scan” menu if there's no “auto” mode.
+		if (!has_auto_mode) {
+			QAction *action = new QAction("Scan", &mode_group);
+			action->setData(QList<QVariant>{"video_mode", 0});
+			mode_submenu.addSeparator();
+			mode_submenu.addAction(action);
+		}
+
+		mode_submenu.setTitle("Input mode");
+		menu.addMenu(&mode_submenu);
+	}
 
 	// --- End of card-dependent choices ---
 
@@ -313,7 +369,7 @@ void GLWidget::show_preview_context_menu(unsigned signal_num, const QPoint &pos)
 	if (global_audio_mixer->get_mapping_mode() == AudioMixer::MappingMode::SIMPLE) {
 		audio_source_action = new QAction("Use as audio source", &menu);
 		audio_source_action->setCheckable(true);
-		if (global_audio_mixer->get_simple_input() == signal_num) {
+		if (global_audio_mixer->get_simple_input() == current_card) {
 			audio_source_action->setChecked(true);
 			audio_source_action->setEnabled(false);
 		}
@@ -335,7 +391,18 @@ void GLWidget::show_preview_context_menu(unsigned signal_num, const QPoint &pos)
 	// Show the menu and look at the result.
 	QAction *selected_item = menu.exec(global_pos);
 	if (audio_source_action != nullptr && selected_item == audio_source_action) {
-		global_audio_mixer->set_simple_input(signal_num);
+		global_audio_mixer->set_simple_input(current_card);
+	} else if (change_url_action != nullptr && selected_item == change_url_action) {
+		// NOTE: We can't use “this” as parent, since the dialog would inherit our style sheet.
+		bool ok;
+		const string url = global_mixer->get_ffmpeg_filename(current_card);
+		QString new_url = QInputDialog::getText(window(), tr("Change URL"),
+			tr("Enter new filename/URL for the given video input:"), QLineEdit::Normal,
+				QString::fromStdString(url), &ok);
+		// FIXME prefill the input
+		if (ok) {
+			global_mixer->set_ffmpeg_filename(current_card, new_url.toStdString());
+		}
 	} else if (selected_item == master_clock_action) {
 		global_mixer->set_master_clock(signal_num);
 	} else if (selected_item != nullptr) {
